"""Publicly accessible classes/routines."""
# from .ensembles import xCanonical, xGrandCanonical
# from .lnpidata import MaskedlnPiDelayed, lnPiArray, lnPiMasked
from . import ensembles, lnpienergy, segment
from .lnpidata import lnPiMasked

# from .lnpienergy import (
#     merge_regions,
#     wFreeEnergy,
#     wFreeEnergyCollection,
#     wFreeEnergyPhases,
# )
from .lnpiseries import lnPiCollection

# from .maskedlnpi_legacy import MaskedlnPiLegacy
from .options import OPTIONS, set_options
from .segment import PhaseCreator

# from .stability import Binodals, Spinodals
# from .utils import dim_to_suffix

# Legacy stuff


# updated versioning scheme
try:
    from importlib.metadata import version as _version
except ImportError:
    # if the fallback library is missing, we are doomed.
    from importlib_metadata import version as _version  # type: ignore[no-redef]


try:
    __version__ = _version("tmmc-lnpy")
except Exception:
    # Local copy or not installed with setuptools.
    # Disable minimum version checks on downstream libraries.
    __version__ = "999"


__author__ = """William P. Krekelberg"""
__email__ = "wpk@nist.gov"


__all__ = [
    "__author__",
    "__email__",
    "__version__",
    # "lnPiArray",
    "lnPiMasked",
    "lnPiCollection",
    "segment",
    "ensembles",
    "lnpienergy",
    # "MaskedlnPiDelayed",
    # "MaskedlnPiLegacy",
    # "xGrandCanonical",
    # "xCanonical",
    # "wFreeEnergyPhases",
    # "wFreeEnergyCollection",
    # "wFreeEnergy",
    # "merge_regions",
    # "Segmenter",
    "PhaseCreator",
    # "peak_local_max_adaptive",
    # "BuildPhases_mu",
    # "BuildPhases_dmu",
    # "dim_to_suffix",
    "set_options",
    "OPTIONS",
    # "Spinodals",
    # "Binodals",
]



"""Common docstrings."""

from module_utilities.docfiller import DocFiller

_docstrings = """
Parameters
----------
copy : bool, optional
    If True, copy the data.  If False, attempt to use view.
copy_kws : mapping, optional
    extra arguments to copy
lnz : float or sequence of float
    Value(s) of ``lnz`` (log of activity)
state_kws : Mapping, optiona
    'State' variables.  Common key/value pairs used by other methods:

    * beta (float) : inverse temperature
    * volume (float) : volume
extra_kws : Mapping, Optional
    Extra parameters.  These are passed along to new objects without modification.
    Common key/value pairs:

* PE (array-like) : value of total potential energy.
    Used in calculation of canonical and grand canonical average potential energy.
fill_value : scalar, default=nan
    Value to fill for masked elements.
data : array-like
    Value of lnPi.  Number of dimensions should be same as ``len(lnz)``.
    For single component, ``data[k]`` is the value of lnPi for `k` particles.
mask_masked | mask : bool or array-like of bool, optional
    Where `mask` is `True`, values are "masked out".
mask_image | mask : bool or array-like of bool, optional
    This uses the "image" convention.  Where `mask` is `True`, values are include.
masks_masked | masks : sequence of array-like of bool
    Value of `masks[i]` is the mask for ith lnPi.
    masks[i] is True where values are "masked out".
mask_convention | convention : string or bool
    Convention for mask.  Allowable values are:

    * 'image' or True : `True` values included, `False` values excluded.
        This is the normal convention in :mod:`scipy.ndimage`.
    * 'masked' or False:  `False` values are included, `True` values are excluded.
        This is the convention in :mod:`numpy.ma`
labels : array-like of int
    Where `labels == k`, indicates regions where data corresponds to the kth label.
base : object
    lnPi data object.
peak_kws : mapping, optional
    Optional parameters to :func:`~lnpy.segment.peak_local_max_adaptive`
watershed_kws : mapping, optional
    Optional parameters to :func:`~skimage.segmentation.watershed`
"""


DOCFILLER_SHARED = DocFiller.from_docstring(_docstrings, combine_keys="parameters")

docfiller_shared = DOCFILLER_SHARED()



"""
Legacy interface to :mod:`lnpy` (:mod:`lnPi`)
=============================================
"""

from warnings import warn

from numpy import VisibleDeprecationWarning

from lnpy import __version__, options, segment, set_options
from lnpy.lnpidata import lnPiMasked as MaskedlnPiDelayed
from lnpy.lnpiseries import lnPiCollection as CollectionlnPi

from . import stability

msg = """\
lnPi is deprecated.  Please transition to using lnpy instead.  Name changes are:

Modules:
lnPi -> lnpy
lnPi.collectionlnpiutils -> lnpy.lnpicollectionutils

Classes:
MaskedlnPiDelayed -> lnPiMasked
CollectionlnPi    -> lnPiCollection
"""


warn(msg, VisibleDeprecationWarning)


__all__ = [
    "MaskedlnPiDelayed",
    "CollectionlnPi",
    "segment",
    "stability",
    "options",
    "set_options",
    "__version__",
]



import json
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import lnpy
import lnpy.stability

path_data = Path(__file__).parent / "water_cluster"


def tag_phases2(x):
    if len(x) > 2:
        raise ValueError("bad tag function")
    argmax0 = np.array([xx.local_argmax()[0] for xx in x])
    return np.where(argmax0 <= x[0].shape[0] / 2, 0, 1)


def load_ref():
    ensemble = json.load(open(path_data / "water_MOF_ensemble.json"))
    lnz = ensemble["betamu0"]
    box = ensemble["box"]
    state_kws = {
        "beta": ensemble["beta"],
        "volume": box[0] * box[1] * box[2],  # being lazy here because the box is cubic
    }

    data = pd.read_csv(path_data / "water_MOF_example.csv")

    # potential energy
    pe = data["U"].values

    # Build the lnPi object
    # DWS Note to self: I'm doing something sloppy here, in that the N values are not specified
    ref = lnpy.lnPiMasked.from_data(
        data=data["lnPi"].values,
        fill_value=np.nan,
        lnz=lnz,
        lnz_data=lnz,
        # state_kws needs to be defined if want to calculate some properties down the road
        state_kws=state_kws,
        # extra_kws is where you pass things which will be passed along to new lnPis, like potential energy
        extra_kws={"PE": pe, "other": pe},
    )
    return ref


@pytest.fixture(params=[0, 1])
def ref(request):
    if request.param == 0:
        return load_ref()
    else:
        import lnpy.examples

        return lnpy.examples.load_example_lnpimasked("watermof")


def get_test_table(o, ref):
    return (
        o.xge.table(
            keys=[
                "betaOmega",
                "nvec",
                "PE",
                "dens",
                "betaF",
                "S",
                "betaG",
                "edge_distance",
            ],
            ref=ref,
        )
        .to_dataframe()
        .reset_index()
    )  # .to_csv('data_0.csv', index=False)


def do_test(phase_creator, ref, fname):
    p = phase_creator.build_phases()

    other = get_test_table(p, ref)
    test = pd.read_csv(path_data / fname)

    pd.testing.assert_frame_equal(other, test)


def test_0(ref):
    phase_creator = lnpy.segment.PhaseCreator(
        nmax=2,  # number of phases
        nmax_peak=10,  # max number of peaks in lnPi
        ref=ref,
        tag_phases=tag_phases2,
        merge_kws={"efac": 0.8},
    )
    do_test(phase_creator, ref, "data_0.csv")


def test_1(ref):
    phase_creator = lnpy.segment.PhaseCreator(
        nmax=20,  # number of phases
        nmax_peak=50,  # max number of peaks in lnPi
        ref=ref,
        tag_phases=None,
        merge_kws={"efac": 0.1},
        segment_kws={"peaks_kws": {"min_distance": [5]}},
    )
    do_test(phase_creator, ref, "data_1.csv")



from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import lnpy

path_data = Path(__file__).parent / "../examples/archived/LJ_mix"


@pytest.fixture
def ref():
    path = path_data / "ljmix4_full.t080.v512.r1.lnpi_o.dat.gz"
    temp = 0.8
    state_kws = {"temp": temp, "beta": 1.0 / temp, "volume": 512}
    lnz = np.array([-2.5, -2.5])

    return (
        lnpy.lnPiMasked.from_table(path, state_kws=state_kws, lnz=lnz).zeromax().pad()
    )


@pytest.fixture
def phase_creator(ref):
    return lnpy.segment.PhaseCreator(
        nmax=2, nmax_peak=4, ref=ref, merge_kws=dict(efac=0.8)
    )


import lnpy.examples


@pytest.fixture(params=[0, 1])
def obj(request, ref, phase_creator):
    if request.param == 0:
        return lnpy.examples.Example(
            ref=ref,
            phase_creator=phase_creator,
            build_phases=phase_creator.build_phases,
        )

    else:
        return lnpy.examples.ljmix_sup_example()


def get_test_table(o, ref):
    return o.xge.table(
        keys=[
            "betaOmega",
            "nvec",
            "PE",
            "dens",
            "betaF",
            "S",
            "betaG",
            "edge_distance",
        ],
        ref=ref,
    )


def test_collection(obj):
    ref, phase_creator = obj.unpack(["ref", "phase_creator"])

    test = pd.read_csv(path_data / "data_0.csv")

    lnz_values = test[["lnz_0", "lnz_1"]].values

    with lnpy.set_options(
        tqdm_leave=True,
        joblib_use=True,
        joblib_len_build=20,
        tqdm_len_build=10,
        tqdm_bar="text",
    ):
        o = lnpy.lnPiCollection.from_builder(
            lnz_values[:], phase_creator.build_phases, unstack=False
        )

    other = get_test_table(o, ref).unstack("sample").to_dataframe().reset_index()

    pd.testing.assert_frame_equal(other, test)



import json
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import lnPi

path_data = Path(__file__).parent / "../examples/archived/LJ_cfs_2.5sig"


def get_lnz(path):
    kB = 1.3806503e-23  # J/K
    kg = 1.66054e-27  # 1/amu
    hPlanck = 6.62606876e-34  # Js

    with open(path) as f:
        metadata = json.load(f)

    deBroglie = (
        hPlanck
        / np.sqrt(
            2.0
            * np.pi
            * (metadata["mass"] * kg)
            * kB
            * (metadata["T*"] * metadata["eps_kB"])
        )
        * 1.0e10
    )  # ang

    mu = metadata["mu*"] + 3.0 * metadata["T*"] * np.log(metadata["sigma"] / deBroglie)
    temp = metadata["T*"]

    lnz = mu / temp

    return lnz, {"beta": 1.0 / temp, "volume": metadata["V*"]}


# get meta data
@pytest.fixture
def ref():
    lnz, state_kws = get_lnz(path_data / "t150.metadata.json")
    print(state_kws)

    # read in potential energy
    pe = pd.read_csv(
        path_data / "ljsf.t150.bulk.v512.r1.energy.dat",
        header=None,
        sep="\\s+",
        names=["n", "e"],
    )["e"].values

    return (
        lnPi.MaskedlnPiDelayed.from_table(
            path_data / "ljsf.t150.bulk.v512.r1.lnpi.dat",
            fill_value=np.nan,
            lnz=lnz,
            # state_kws needs to be defined if want to calculate some properties
            # down the road
            state_kws=state_kws,
            # extra_kws is where you pass things which will be passed along to
            # new lnPis, like potential energy
            extra_kws={"PE": pe},
        )
        .zeromax()
        .pad()
    )


@pytest.fixture
def phase_creator(ref):
    return lnPi.segment.PhaseCreator(
        nmax=1,
        nmax_peak=4,
        ref=ref,
        merge_kws=dict(efac=0.8),
        # if want to id phases based on some
        # callable, then set this.
        # the callable is passes a list of
        # MaskedlnPi objects, and should return
        # a list of phaseid
        tag_phases=None
        # tag_phases=tag_phases2
    )


@pytest.fixture
def build_phases(phase_creator):
    # choose a particular factory method for creating phases
    return phase_creator.build_phases_mu([None])


import lnpy.examples


# can drop param = 0 after make sure all good
@pytest.fixture(params=[1])
def obj(request, ref, phase_creator, build_phases):
    if request.param == 0:
        return lnpy.examples.Example(
            ref=ref, phase_creator=phase_creator, build_phases=build_phases
        )

    else:
        return lnpy.examples.lj_sup_example()


@pytest.fixture
def test_table():
    table = pd.read_csv(path_data / "data_0.csv")
    return table


def get_test_table(o, ref):
    return (
        o.xge.table(
            keys=[
                "betaOmega",
                "nvec",
                "PE",
                "dens",
                "betaF",
                "S",
                "betaG",
                "edge_distance",
            ],
            ref=ref,
        )
        .to_dataframe()
        .reset_index()
    )  # .to_csv('data_0.csv', index=False)


def test_collection_properties(obj, test_table):
    ref, build_phases = obj.unpack(["ref", "build_phases"])

    # for big builds, take advantage of progress bar, and parallel builds
    lnzs = np.linspace(-10, 3.5, 2000)

    # by default, progress bar hides itself after completion.  use context manager to keep it
    # note that for this example (where only have a single phase), doesn't really make a difference
    with lnPi.set_options(tqdm_leave=True, joblib_use=False, tqdm_bar="text"):
        o = lnPi.CollectionlnPi.from_builder(lnzs, build_phases)

    other = get_test_table(o, ref)

    pd.testing.assert_frame_equal(test_table, other)


@pytest.fixture
def test_table_can():
    table = pd.read_csv(path_data / "data_0_can.csv")

    return table


def get_test_table_can(ref):
    return (
        ref.xce.table(
            keys=["S", "Z", "betaE", "betaF", "betaOmega", "betamu", "dens", "ntot"]
        )
        .to_dataframe()
        .reset_index()
    )


def test_canonical_properties(obj, test_table_can):
    ref = obj.ref
    other = get_test_table_can(ref)
    pd.testing.assert_frame_equal(other, test_table_can)


def test_nice_grid(obj):
    ref, build_phases = obj.unpack(["ref", "build_phases"])
    import lnPi.collectionlnpiutils

    with lnPi.set_options(joblib_use=True):
        o_course, o = lnPi.collectionlnpiutils.limited_collection(
            build_phases,
            dlnz=0.01,
            offsets=[-10, +10],
            even_grid=True,  # but lnzs on same grid as dlnz
            digits=2,  # round lnzs to this number of digits
            edge_distance_min=10,
            dens_min=0.001,
        )

    other_course = get_test_table(o_course, ref)
    test_course = pd.read_csv(path_data / "data_0_course.csv")
    pd.testing.assert_frame_equal(other_course, test_course)

    other_fine = get_test_table(o, ref)
    test = pd.read_csv(path_data / "data_0_fine.csv")
    pd.testing.assert_frame_equal(other_fine, test)


# def test_deprecation():
#     from numpy import VisibleDeprecationWarning
#     with pytest.warns(VisibleDeprecationWarning):
#         import lnPi



from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import lnpy
import lnpy.stability

path_data = Path(__file__).parent / "../examples/archived/HS_mix"


# function to tag 'LD' and 'HD' phases
def tag_phases2(x):
    if len(x) > 2:
        raise ValueError("bad tag function")
    argmax0 = np.array([xx.local_argmax()[0] for xx in x])
    return np.where(argmax0 <= x[0].shape[0] / 2, 0, 1)


@pytest.fixture
def ref():
    return (
        lnpy.lnPiMasked.from_table(
            path_data / "nahs_asym_mix.07_07_07.r1.lnpi_o.dat",
            lnz=np.array([0.5, 0.5]),
            state_kws={"beta": 1.0, "volume": 1.0},
        )
        .zeromax()
        .pad()
    )


@pytest.fixture
def phase_creator(ref):
    return lnpy.segment.PhaseCreator(
        nmax=2,
        nmax_peak=4,
        ref=ref,
        tag_phases=tag_phases2,
        merge_kws=dict(efac=0.8),
    )


import lnpy.examples


@pytest.fixture(params=[0, 1])
def obj(request, ref, phase_creator):
    if request.param == 0:
        return lnpy.examples.Example(
            ref=ref, phase_creator=phase_creator, build_phases=None
        )
    else:
        return lnpy.examples.hsmix_example()


@pytest.fixture
def lnz2():
    return 0.5


@pytest.fixture
def lnzs():
    return np.linspace(-8, 8, 20)


@pytest.fixture
def build_phases(obj, lnz2):
    return obj.phase_creator.build_phases_mu([None, lnz2])


def get_test_table(o, ref):
    return (
        o.xge.table(
            keys=[
                "betaOmega",
                "nvec",
                "PE",
                "dens",
                "betaF",
                "S",
                "betaG",
                "edge_distance",
            ],
            ref=ref,
        )
        .to_dataframe()
        .reset_index()
    )  # .to_csv('data_0.csv', index=False)


def test_collection(obj, build_phases, lnzs):
    ref = obj.ref

    c = lnpy.lnPiCollection.from_builder(lnzs, build_phases)
    c.spinodal(2, build_phases, inplace=True, unstack=True)
    c.binodal(2, build_phases, inplace=True, unstack=True)

    for path, obj in [
        ("data_0", c),
        ("data_0_spin", c.spinodal.access),
        ("data_0_bino", c.binodal.access),
    ]:
        test = pd.read_csv(path_data / (path + ".csv"))
        other = get_test_table(obj, ref)

        pd.testing.assert_frame_equal(test, other)

        test = pd.read_csv(path_data / (path + "_dw.csv"))
        other = obj.wfe.dw.to_frame().reset_index()

        pd.testing.assert_frame_equal(test, other)
